/*
GNU General Public License

ArchichronosApplet - Modern Art Clock Applet
Copyright (C) 2000  Howard A Kistler

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package com.hexidec.tool;

import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Calendar;
import java.util.Random;

public class ArchichronosApplet extends Applet implements Runnable {

	private Thread thrTimer;
	private Canvas cnvClock;
	private Image  bufferAlpha;
	private Image  imageBack;
	private Image  imageFront;
	private int    oldTime = -1;

	private int appWidth  = 200;
	private int appHeight = 200;
	private int discWidth = 10;

	// Display Flags
	private boolean showHandHour    = true;
	private boolean showHandMinute  = true;
	private boolean showHandSecond  = true;
	private boolean showNumericTime = true;
	private boolean showAsRings     = true;
	private boolean showFasterInner = true;
	private boolean showArcTrim     = false;

	// Default Colors
	private Color colorBack   = Color.black;
	private Color colorTrim   = Color.white;
	private Color colorHour   = Color.blue;
	private Color colorMinute = Color.yellow;
	private Color colorSecond = Color.gray;
	private Color colorTime   = Color.red;

	// Random Colors Flags
	private boolean randomHour   = false;
	private boolean randomMinute = false;
	private boolean randomSecond = false;

	// Cycle Colors Flags
	private boolean cycleHour   = false;
	private boolean cycleMinute = false;
	private boolean cycleSecond = false;

	// Color Cycles
	private Color[] colorCycleHour =
	{
		new Color(0, 0, 0), // 0 hour, unused
		new Color(128,   0,  64),
		new Color(128,   0, 128),
		new Color(128,   0, 192),
		new Color(128,  64, 128),
		new Color(128,  64,  64),
		new Color(128, 128,   0),
		new Color(192, 192,  64),
		new Color(255, 255, 128),
		new Color(255, 192, 128),
		new Color(255, 128, 128),
		new Color(192,  64,  64),
		new Color(128,   0,   0)
	};

	private Color[] colorCycleMinute =
	{
		new Color(  0, 255,   0),
		new Color(  0, 245,   5),
		new Color(  0, 235,   5),
		new Color(  0, 225,  15),
		new Color(  0, 215,  15),
		new Color(  0, 205,  25),
		new Color(  0, 195,  25),
		new Color(  0, 185,  35),
		new Color(  0, 175,  45),
		new Color(  0, 165,  55),
		new Color(  0, 155,  65),
		new Color(  0, 145,  75),
		new Color(  0, 135,  85),
		new Color(  0, 125,  95),
		new Color(  0, 115, 105),
		new Color(  0, 105, 115),
		new Color(  0,  95, 125),
		new Color(  0,  85, 135),
		new Color(  0,  75, 145),
		new Color(  0,  65, 155),
		new Color(  0,  55, 165),
		new Color(  0,  45, 175),
		new Color(  0,  35, 185),
		new Color(  0,  25, 195),
		new Color(  0,  25, 205),
		new Color(  0,  15, 215),
		new Color(  0,  15, 225),
		new Color(  0,   5, 235),
		new Color(  0,   5, 245),
		new Color(  0,   0, 255),
		new Color(  0,   0, 255),
		new Color(  0,   5, 255),
		new Color(  0,   5, 255),
		new Color(  0,  15, 255),
		new Color(  0,  15, 255),
		new Color(  0,  25, 255),
		new Color(  0,  25, 255),
		new Color(  0,  35, 255),
		new Color(  0,  45, 255),
		new Color(  0,  55, 255),
		new Color(  0,  65, 255),
		new Color(  0,  75, 255),
		new Color(  0,  85, 255),
		new Color(  0,  95, 255),
		new Color(  0, 105, 255),
		new Color(  0, 115, 255),
		new Color(  0, 125, 255),
		new Color(  0, 135, 255),
		new Color(  0, 145, 255),
		new Color(  0, 155, 255),
		new Color(  0, 165, 255),
		new Color(  0, 175, 255),
		new Color(  0, 185, 255),
		new Color(  0, 195, 255),
		new Color(  0, 205, 255),
		new Color(  0, 215, 255),
		new Color(  0, 225, 255),
		new Color(  0, 235, 255),
		new Color(  0, 245, 255),
		new Color(  0, 255, 255)
	};

	private int orderHour   = 0;
	private int orderMinute = 1;
	private int orderSecond = 2;
	private int orderCenter = 3;

	public void init() {
		appWidth  = this.getWidth();
		appHeight = this.getHeight();

		cnvClock = new Canvas();
		cnvClock.setSize(appWidth, appHeight);
		cnvClock.setBounds(0, 0, appWidth, appHeight);

		bufferAlpha = createImage(appWidth, appHeight);

		String appHandHour     = getParameter("SHOWHANDHOUR");
		String appHandMinute   = getParameter("SHOWHANDMINUTE");
		String appHandSecond   = getParameter("SHOWHANDSECOND");
		String appNumericTime  = getParameter("SHOWNUMERICTIME");
		String appTimeAsRings  = getParameter("SHOWTIMEASRINGS");
		String appFasterInner  = getParameter("SHOWFASTERINNER");
		String appArcTrim      = getParameter("SHOWARCTRIM");
		String appDiscWidth    = getParameter("DISCWIDTH");
		String appColorBack    = getParameter("COLORBACK");
		String appColorTrim    = getParameter("COLORTRIM");
		String appColorHour    = getParameter("COLORHOUR");
		String appColorMinute  = getParameter("COLORMINUTE");
		String appColorSecond  = getParameter("COLORSECOND");
		String appColorTime    = getParameter("COLORTIME");
		String appCycleHour    = getParameter("CYCLECOLORHOUR");
		String appCycleMinute  = getParameter("CYCLECOLORMINUTE");
		String appCycleSecond  = getParameter("CYCLECOLORSECOND");
		String appRandomHour   = getParameter("RANDOMCOLORHOUR");
		String appRandomMinute = getParameter("RANDOMCOLORMINUTE");
		String appRandomSecond = getParameter("RANDOMCOLORSECOND");
		String appImageBack    = getParameter("IMAGEBACK");
		String appImageFront   = getParameter("IMAGEFRONT");

		if(appHandHour     != null) { showHandHour    = (appHandHour.equalsIgnoreCase("YES")); }
		if(appHandMinute   != null) { showHandMinute  = (appHandMinute.equalsIgnoreCase("YES")); }
		if(appHandSecond   != null) { showHandSecond  = (appHandSecond.equalsIgnoreCase("YES")); }
		if(appNumericTime  != null) { showNumericTime = (appNumericTime.equalsIgnoreCase("YES")); }
		if(appTimeAsRings  != null) { showAsRings     = (appTimeAsRings.equalsIgnoreCase("YES")); }
		if(appFasterInner  != null) { showFasterInner = (appFasterInner.equalsIgnoreCase("YES")); }
		if(appArcTrim      != null) { showArcTrim     = (appArcTrim.equalsIgnoreCase("YES")); }
		if(appDiscWidth    != null) { discWidth       = Integer.parseInt(appDiscWidth); }
		if(appColorBack    != null) { colorBack       = parseColorCode(appColorBack); }
		if(appColorTrim    != null) { colorTrim       = parseColorCode(appColorTrim); }
		if(appColorHour    != null) { colorHour       = parseColorCode(appColorHour); }
		if(appColorMinute  != null) { colorMinute     = parseColorCode(appColorMinute); }
		if(appColorSecond  != null) { colorSecond     = parseColorCode(appColorSecond); }
		if(appColorTime    != null) { colorTime       = parseColorCode(appColorTime); }
		if(appCycleHour    != null) { cycleHour       = (appCycleHour.equalsIgnoreCase("YES")); }
		if(appCycleMinute  != null) { cycleMinute     = (appCycleMinute.equalsIgnoreCase("YES")); }
		if(appCycleSecond  != null) { cycleSecond     = (appCycleSecond.equalsIgnoreCase("YES")); }
		if(appRandomHour   != null) { randomHour      = (appRandomHour.equalsIgnoreCase("YES")); }
		if(appRandomMinute != null) { randomMinute    = (appRandomMinute.equalsIgnoreCase("YES")); }
		if(appRandomSecond != null) { randomSecond    = (appRandomSecond.equalsIgnoreCase("YES")); }

		if(appImageBack  != null) { imageBack  = Toolkit.getDefaultToolkit().getImage(appImageBack); }
		if(appImageFront != null) { imageFront = Toolkit.getDefaultToolkit().getImage(appImageFront); }

		// work out the order (and therefore the sizing) of the hands
		if(showFasterInner)
		{ // hour-to-second, heading inward
			int currLevel = 0;
			if(showHandHour)
			{
				orderHour = currLevel;
				currLevel++;
			}
			if(showHandMinute)
			{
				orderMinute = currLevel;
				currLevel++;
			}
			if(showHandSecond)
			{
				orderSecond = currLevel;
				currLevel++;
			}
			orderCenter = currLevel;
		}
		else
		{ // second-to-hour, heading inward
			int currLevel = 0;
			if(showHandSecond)
			{
				orderSecond = currLevel;
				currLevel++;
			}
			if(showHandMinute)
			{
				orderMinute = currLevel;
				currLevel++;
			}
			if(showHandHour)
			{
				orderHour = currLevel;
				currLevel++;
			}
			orderCenter = currLevel;
		}

		this.setLayout(new BorderLayout());
		this.add(cnvClock, BorderLayout.CENTER);
	}

	public void start()
	{
		if(thrTimer == null)
		{
			thrTimer = new Thread(this);
		}
		thrTimer.start();
	}

	public void stop()
	{
		thrTimer = null;
	}

	public void destroy()
	{
		super.destroy();
	}

	public void run()
	{
		while(true)
		{
			if(oldTime != Calendar.getInstance().get(Calendar.SECOND))
			{
				doClock();
				oldTime = Calendar.getInstance().get(Calendar.SECOND);
			}
		}
	}

	private void paintHand(Color paintColor, int arcStartX, int arcStartY, int arcWidth, int arcHeight, int arcAngle)
	{
		Graphics graphicsObject = bufferAlpha.getGraphics();
		graphicsObject.setColor(paintColor);
		graphicsObject.fillArc(arcStartX, arcStartY, arcWidth, arcHeight, 90, -arcAngle);
		if(showArcTrim)
		{
			graphicsObject.setColor(colorTrim);
			graphicsObject.drawArc(arcStartX, arcStartY, arcWidth, arcHeight, 90, -arcAngle);
		}
	}

	private Color subtleColorChange(Color color)
	{
		Random rnd = new Random(System.currentTimeMillis());
		int which = rnd.nextInt(3);
		int offset = rnd.nextInt(2);
		if(offset == 0) { offset = -1; }
		int colorR = color.getRed();
		int colorG = color.getGreen();
		int colorB = color.getBlue();
		if     (which == 0) { colorR = colorR + offset; }
		else if(which == 1) { colorG = colorG + offset; }
		else if(which == 2) { colorB = colorB + offset; }
		if(colorR < 0)   { colorR = 0; }
		if(colorG < 0)   { colorG = 0; }
		if(colorB < 0)   { colorB = 0; }
		if(colorR > 255) { colorR = 255; }
		if(colorG > 255) { colorG = 255; }
		if(colorB > 255) { colorB = 255; }
		return new Color(colorR, colorG, colorB);
	}

	private void doClock()
	{
		if(bufferAlpha == null)
		{
			bufferAlpha = createImage(appWidth, appHeight);
		}
		Calendar calen = Calendar.getInstance();
		Graphics g2D = bufferAlpha.getGraphics();
		Graphics gCanvas = cnvClock.getGraphics();

		// clear background
		g2D.setColor(colorBack);
		g2D.fillRect(0, 0, appWidth, appHeight);

		// paint background image, if present
		if(imageBack != null)
		{
			g2D.drawImage(imageBack, 0, 0, this);
		}

		int calenHour12 = calen.get(Calendar.HOUR);
		if(calenHour12 == 0) { calenHour12 = 12; }

		if(showFasterInner)
		{
			if(showHandHour)
			{
				if(cycleHour)
				{
					colorHour = colorCycleHour[calenHour12];
				}
				else if(randomHour)
				{
					colorHour = subtleColorChange(colorHour);
				}
				paintHand(colorHour, discWidth * orderHour, discWidth * orderHour, appWidth - (discWidth * (orderHour * 2)), appHeight - (discWidth * (orderHour * 2)), (calenHour12 * 30));
			}
			if(showHandMinute)
			{
				if(cycleMinute)
				{
					colorMinute = colorCycleMinute[calen.get(Calendar.MINUTE)];
				}
				else if(randomMinute)
				{
					colorMinute = subtleColorChange(colorMinute);
				}
				paintHand(colorMinute, discWidth * orderMinute, discWidth * orderMinute, appWidth - (discWidth * (orderMinute * 2)), appHeight - (discWidth * (orderMinute * 2)), (calen.get(Calendar.MINUTE) * 6));
			}
			if(showHandSecond)
			{
				if(cycleSecond)
				{
//					colorSecond = colorCycleSecond[calen.get(Calendar.SECOND)];
				}
				else if(randomSecond)
				{
					colorSecond = subtleColorChange(colorSecond);
				}
				paintHand(colorSecond, discWidth * orderSecond, discWidth * orderSecond, appWidth - (discWidth * (orderSecond * 2)), appHeight - (discWidth * (orderSecond * 2)), (calen.get(Calendar.SECOND) * 6));
			}
		}
		else
		{
			if(showHandSecond)
			{
				if(cycleSecond)
				{
//					colorSecond = colorCycleSecond[calen.get(Calendar.SECOND)];
				}
				else if(randomSecond)
				{
					colorSecond = subtleColorChange(colorSecond);
				}
				paintHand(colorSecond, discWidth * orderSecond, discWidth * orderSecond, appWidth - (discWidth * (orderSecond * 2)), appHeight - (discWidth * (orderSecond * 2)), (calen.get(Calendar.SECOND) * 6));
			}
			if(showHandMinute)
			{
				if(cycleMinute)
				{
					colorMinute = colorCycleMinute[calen.get(Calendar.MINUTE)];
				}
				else if(randomMinute)
				{
					colorMinute = subtleColorChange(colorMinute);
				}
				paintHand(colorMinute, discWidth * orderMinute, discWidth * orderMinute, appWidth - (discWidth * (orderMinute * 2)), appHeight - (discWidth * (orderMinute * 2)), (calen.get(Calendar.MINUTE) * 6));
			}
			if(showHandHour)
			{
				if(cycleHour)
				{
					colorHour = colorCycleHour[calenHour12];
				}
				else if(randomHour)
				{
					colorHour = subtleColorChange(colorHour);
				}
				paintHand(colorHour, discWidth * orderHour, discWidth * orderHour, appWidth - (discWidth * (orderHour * 2)), appHeight - (discWidth * (orderHour * 2)), (calenHour12 * 30));
			}
		}

		// paint inner disc
		if(showAsRings)
		{
			g2D.setColor(colorBack);
			g2D.fillArc(discWidth * orderCenter, discWidth * orderCenter, appWidth - (discWidth * (orderCenter * 2)), appHeight - (discWidth * (orderCenter * 2)), 90, 360);
		}

		// paint foreground image, if present
		if(imageFront != null)
		{
			g2D.drawImage(imageFront, 0, 0, this);
		}

		// paint time string
		if(showNumericTime)
		{
			g2D.setColor(colorTime);
			String strTime = (calenHour12 < 10 ? "0" : "") + calenHour12 + ":" + (calen.get(Calendar.MINUTE) < 10 ? "0" : "") + calen.get(Calendar.MINUTE) + (showHandSecond ? ":" + (calen.get(Calendar.SECOND) < 10 ? "0" : "") + calen.get(Calendar.SECOND) : "");
			g2D.setFont(new Font("monospaced", Font.BOLD, 14));
			g2D.drawString(strTime, (appWidth / 2) - 32, (appHeight / 2) + 4);
		}

		// paint buffered image to visible canvas
		gCanvas.drawImage(bufferAlpha, 0, 0, this);
	}

	private Color parseColorCode(String strColor)
	{
		if((strColor.indexOf("r") < 0) || (strColor.indexOf("g") < 0) || (strColor.indexOf("b") < 0))
		{
			return Color.black;
		}
		int rVal = Integer.parseInt(strColor.substring(strColor.indexOf("r") + 1, strColor.indexOf("g")));
		int gVal = Integer.parseInt(strColor.substring(strColor.indexOf("g") + 1, strColor.indexOf("b")));
		int bVal = Integer.parseInt(strColor.substring(strColor.indexOf("b") + 1, strColor.length()));
		return new Color(rVal, gVal, bVal);
	}

	public void paint(Graphics g)
	{
		super.paint(g);
		if(bufferAlpha != null)
		{
			Graphics gCanvas = cnvClock.getGraphics();
			gCanvas.drawImage(bufferAlpha, 0, 0, this);
		}
	}

}
